@@ -3,7 +3,7 @@ typedload
 Module to load data into typed data structures
 """
 
-# Copyright (C) 2018-2023 Salvo "LtWorf" Tomaselli
+# Copyright (C) 2018-2024 Salvo "LtWorf" Tomaselli
 #
 # typedload is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -238,9 +238,11 @@ class Loader:
 
         self._indexcache = {}  # type: Dict[Any, Callable[[Loader, Any, Any], Any]]
 
-        self._objfieldscache = {}  # type: Dict[Type, Tuple[Set[str], Set[str], Dict[str, Type], Dict[str, str]]]
+        self._objfieldscache = {}  # type: Dict[Type, Any]
 
-        self._unionload_discriminatorcache = {}  # type: Dict[Type, Tuple[Optional[str], Optional[Dict[Any, Type]]]]
+        self._unionload_discriminatorcache = {}  # type: Dict[int, Tuple[Optional[str], Optional[Dict[Any, Type]]]]
+
+        self._union_sorted_args_cache = {}  # type: Dict[int, List[Type]]
 
     def index(self, type_: Type[T]) -> int:
         """
@@ -260,6 +262,15 @@ class Loader:
                 return i
         raise ValueError('No matching condition found')
 
+    def _index_and_cache(self, type_: Type[T]) -> int:
+        """
+        Same as index() but the result is added to the cache
+        """
+        r = self.index(type_)
+        self._indexcache[type_] = self.handlers[r][1]
+        return r
+
+
     def load(self, value: Any, type_: Type[T], *, annotation: Optional[Annotation] = None) -> T:
         """
         Loads value into the typed data structure.
@@ -374,10 +385,9 @@ def _dictload(l: Loader, value: Any, type_) -> Dict:
     key_type_basic = key_type in l.basictypes
     value_type_basic = value_type in l.basictypes
 
-    key_handler = l._indexcache.get(key_type)
-    if key_handler is not None:
-        key_f = key_handler
-    else:
+    try:
+        key_f = l._indexcache[key_type]
+    except KeyError:
         try:
             key_f = l._indexcache[key_type] = l.handlers[l.index(key_type)][1]
         except ValueError:
@@ -388,12 +398,11 @@ def _dictload(l: Loader, value: Any, type_) -> Dict:
             )
 
     # Same thing for the value
-    value_handler = l._indexcache.get(value_type)
-    if value_handler is not None:
-        value_f = value_handler
-    else:
+    try:
+        value_f = l._indexcache[value_type]
+    except KeyError:
         try:
-            key_f = l._indexcache[value_type] = l.handlers[l.index(value_type)][1]
+            value_f = l._indexcache[value_type] = l.handlers[l.index(value_type)][1]
         except ValueError:
             raise TypedloadValueError(
                 'Cannot deal with value of type %s (value of %s)' % (tname(value_type), tname(type_)),
@@ -428,24 +437,27 @@ def _tupleload(l: Loader, value: Any, type_) -> Tuple:
     """
     args = type_.__args__
 
-    if len(args) == 2 and args[1] == ...: # Tuple[something, ...]
+    len_args = len(args)
+
+    if len_args == 2 and args[1] == ...: # Tuple[something, ...]
         return _iterload(l, value, type_, tuple)
 
     # Tuple[something, something, somethingelse]
     if isinstance(value, dict):
         raise TypedloadTypeError('Unable to load dictionary as a tuple', value=value, type_=type_)
-    if l.failonextra and len(value) > len(args):
+
+    len_value = len(value)
+    if l.failonextra and len_value > len_args:
         raise TypedloadValueError('Value is too long for type %s' % tname(type_), value=value, type_=type_)
-    elif len(value) < len(args):
+    elif len_value < len_args:
         raise TypedloadValueError('Value is too short for type %s' % tname(type_), value=value, type_=type_)
 
     ctr = count(1)  # Keep track of the position in the tuple
 
     try:
-        return tuple(v if t in l.basictypes and type(v) == t else h(l, v, t)
-            for v, h, t in zip(
+        return tuple(v if t in l.basictypes and type(v) == t else (l._indexcache.get(t) or l.handlers[l._index_and_cache(t)][1])(l, v, t)
+            for v, t in zip(
                 compress(value, ctr),
-                (l._indexcache.get(t) or l.handlers[l.index(t)][1] for t in args),
                 args
             )
         )
@@ -491,10 +503,9 @@ def _dataclassload(l: Loader, value: Dict[str, Any], type_) -> Any:
     """
     This loads a Dict[str, Any] into a NamedTuple.
     """
-    cached = l._objfieldscache.get(type_)
-    if cached:
-        fields, necessary_fields, type_hints, transforms = cached
-    else:
+    try:
+        fields, necessary_fields, type_hints, transforms = l._objfieldscache[type_]
+    except KeyError:
         fields = set(type_.__dataclass_fields__.keys())
         necessary_fields = {k for k,v in type_.__dataclass_fields__.items() if
                             v.init == True and # Is a field for the constructor
@@ -516,14 +527,15 @@ def _dataclassload(l: Loader, value: Dict[str, Any], type_) -> Any:
                     transforms[name] = pyname
         l._objfieldscache[type_] = (fields, necessary_fields, type_hints, transforms)
 
-    try:
-        value = _mangle_names(transforms, value, l.failonextra)
-    except ValueError as e:
-        raise TypedloadValueError(str(e), value=value, type_=type_)
-    except AttributeError as e:
-        raise TypedloadAttributeError(str(e), value=value, type_=type_)
+    if transforms:
+        try:
+            value = _mangle_names(transforms, value, l.failonextra)
+        except ValueError as e:
+            raise TypedloadValueError(str(e), value=value, type_=type_)
+        except AttributeError as e:
+            raise TypedloadAttributeError(str(e), value=value, type_=type_)
 
-    return _objloader(l, fields, necessary_fields, type_hints, value, type_)
+    return _objloader(l, fields, necessary_fields, type_hints, value, type_, False)
 
 
 def _dictequivalence(l: Loader, value: Any) -> Any:
@@ -542,23 +554,24 @@ def _dictequivalence(l: Loader, value: Any) -> Any:
     return value
 
 
-def _objloader(l: Loader, fields: Set[str], necessary_fields: Set[str], type_hints, value: Any, type_) -> Any:
+def _objloader(l: Loader, fields: Set[str], necessary_fields: Set[str], type_hints, value: Any, type_, early_needed_fields_check: bool) -> Any:
     '''
     Helper function to load dict-like data into an object.
     '''
-    try:
-        vfields = set(value.keys())
-    except AttributeError as e:
+
+    if l.dictequivalence and not isinstance(value, dict):
         newvalue = _dictequivalence(l, value)
 
         if newvalue is value:
-            raise TypedloadAttributeError(str(e), value=value, type_=type_)
+            raise TypedloadAttributeError('Not a dictionary: %s' % value, value=value, type_=type_)
         else:
             value = newvalue
-            vfields = set(value.keys())
 
-    if len(necessary_fields.intersection(vfields)) != len(necessary_fields):
-        raise TypedloadValueError(
+    if early_needed_fields_check or l.failonextra:
+        vfields = set(value.keys())
+
+    if early_needed_fields_check and len(necessary_fields.intersection(vfields)) != len(necessary_fields):
+        raise TypedloadTypeError(
             'Value does not contain fields: %s which are necessary for type %s' % (
                 necessary_fields.difference(vfields),
                 tname(type_)
@@ -567,9 +580,10 @@ def _objloader(l: Loader, fields: Set[str], necessary_fields: Set[str], type_hin
             type_=type_,
         )
 
+
     if l.failonextra and len(extra_fields := vfields.difference(fields)):
         extra = ', '.join(extra_fields)
-        raise TypedloadValueError(
+        raise TypedloadTypeError(
             'Dictionary has unrecognized fields: %s and cannot be loaded into %s' % (extra, tname(type_)),
             value=value,
             type_=type_,
@@ -583,10 +597,9 @@ def _objloader(l: Loader, fields: Set[str], necessary_fields: Set[str], type_hin
 
         # loading field directly, skipping load()
         field_type = type_hints[k]
-        cached_loader = l._indexcache.get(field_type)
-        if cached_loader:
-            loader_f = cached_loader
-        else:
+        try:
+            loader_f = l._indexcache[field_type]
+        except KeyError:
             try:
                 loader_f = l._indexcache[field_type] = l.handlers[l.index(field_type)][1]
             except ValueError:
@@ -604,6 +617,17 @@ def _objloader(l: Loader, fields: Set[str], necessary_fields: Set[str], type_hin
     try:
         return type_(**params)
     except TypeError as e:
+        vfields = set(value.keys())
+        # Check the necessary fields only if there was an error loading
+        if len(necessary_fields.intersection(vfields)) != len(necessary_fields):
+            raise TypedloadTypeError(
+                'Value does not contain fields: %s which are necessary for type %s' % (
+                    necessary_fields.difference(vfields),
+                    tname(type_)
+                ),
+                value=value,
+                type_=type_,
+            )
         raise TypedloadTypeError(e)
     except ValueError as e:
         raise TypedloadValueError(e)
@@ -613,15 +637,21 @@ def _namedtupleload(l: Loader, value: Any, type_) -> Any:
     """
     This loads a Dict[str, Any] into a NamedTuple.
     """
-    if l.pep563:
-        type_hints = get_type_hints(type_)
-    else:
-        type_hints = type_.__annotations__
-    fields = set(type_hints.keys())
-    optional_fields = set(getattr(type_, '_field_defaults', {}).keys())
-    necessary_fields = fields.difference(optional_fields)
 
-    return _objloader(l, fields, necessary_fields, type_hints, value, type_)
+    try:
+        fields, necessary_fields, type_hints = l._objfieldscache[type_]
+    except KeyError:
+        if l.pep563:
+            type_hints = get_type_hints(type_)
+        else:
+            type_hints = type_.__annotations__
+        fields = set(type_hints.keys())
+        optional_fields = set(getattr(type_, '_field_defaults', {}).keys())
+        necessary_fields = fields.difference(optional_fields)
+
+        l._objfieldscache[type_] = fields, necessary_fields, type_hints
+
+    return _objloader(l, fields, necessary_fields, type_hints, value, type_, False)
 
 
 def _typeddictload(l: Loader, value: Any, type_) -> Any:
@@ -648,7 +678,7 @@ def _typeddictload(l: Loader, value: Any, type_) -> Any:
             type_hints[k] = notrequiredtype(v)
             necessary_fields.discard(k)
 
-    return _objloader(l, fields, necessary_fields, type_hints, value, type_)
+    return _objloader(l, fields, necessary_fields, type_hints, value, type_, True)
 
 
 def _unionload(l: Loader, value: Any, type_) -> Any:
@@ -664,6 +694,7 @@ def _unionload(l: Loader, value: Any, type_) -> Any:
     args = uniontypes(type_)
 
     value_type = type(value)
+    type_id = id(type_)
 
     # Do not convert basic types, if possible
     if value_type in l.basictypes and value_type in args:
@@ -672,17 +703,23 @@ def _unionload(l: Loader, value: Any, type_) -> Any:
     exceptions = []
 
     # Give a score to the types
-    sorted_args = list(args)  # type: List[Type]
-    sorted_args.sort(key=lambda i: i in l.basictypes)
+    try:
+        sorted_args = l._union_sorted_args_cache[type_id]  # type: List[Type]
+    except KeyError:
+        sorted_args = list(args)
+        sorted_args.sort(key=lambda i: i in l.basictypes)
+        l._union_sorted_args_cache[type_id] = sorted_args
 
     # For object types, bump up the type whose literal is matching
     if hasattr(value, 'get'):
         # Seems we have an object
         # Bump up if the Literal field matches
-        discriminatorscache = l._unionload_discriminatorcache.get(type_)  # type: Optional[Tuple[Optional[str], Optional[Dict[Any, Type]]]]
+
+        try:
+            discriminatorscache = l._unionload_discriminatorcache[type_id]  # type: Tuple[Optional[str], Optional[Dict[Any, Type]]]
 
         # First time generate the deep inspection for literal
-        if discriminatorscache is None:
+        except KeyError:
             # type → {key: valueset}
             data = {t: discriminatorliterals(t) for t in args}
             # shared keys that have literals in every object of the union
@@ -697,7 +734,7 @@ def _unionload(l: Loader, value: Any, type_) -> Any:
                 discriminatorscache = key, cachedict
             else:
                 discriminatorscache = None, None
-            l._unionload_discriminatorcache[type_] = discriminatorscache
+            l._unionload_discriminatorcache[type_id] = discriminatorscache
 
         # Cache is created, use it
         # It's a tuple key, {value: type}
@@ -714,10 +751,9 @@ def _unionload(l: Loader, value: Any, type_) -> Any:
     for t in sorted_args:
         try:
             # Skip calling load()
-            handler = l._indexcache.get(t)
-            if handler is not None:
-                f = handler
-            else:
+            try:
+                f = l._indexcache[t]
+            except KeyError:
                 try:
                     f = l._indexcache[t] = l.handlers[l.index(t)][1]
                 except ValueError:
@@ -855,29 +891,35 @@ def _get_attr_converter_type(c: "Callable"):
 
 
 def _attrload(l: Loader, value: Any, type_) -> Any:
-    from attr._make import _Nothing as NOTHING
-
-    fields = {i.name for i in type_.__attrs_attrs__}
-    necessary_fields = set()
-    type_hints = {i.name: (_get_attr_converter_type(i.converter) if i.converter else i.type) for i in type_.__attrs_attrs__}
-    namesmap = {}  # type: Dict[str, str]
-
-    for attribute in type_.__attrs_attrs__:
-        if attribute.default is NOTHING and attribute.init:
-            necessary_fields.add(attribute.name)
-
-        # Manage name mangling
-        if l.mangle_key in attribute.metadata:
-            namesmap[attribute.metadata[l.mangle_key]] = attribute.name
 
     try:
-        value = _mangle_names(namesmap, value, l.failonextra)
-    except ValueError as e:
-        raise TypedloadValueError(str(e), value=value, type_=type_)
-    except AttributeError as e:
-        raise TypedloadAttributeError(str(e), value=value, type_=type_)
+        fields, necessary_fields, type_hints, namesmap = l._objfieldscache[type_]
+    except KeyError:
+        from attr._make import _Nothing as NOTHING
 
-    return _objloader(l, fields, necessary_fields, type_hints, value, type_)
+        fields = {i.name for i in type_.__attrs_attrs__}
+        necessary_fields = set()
+        type_hints = {i.name: (_get_attr_converter_type(i.converter) if i.converter else i.type) for i in type_.__attrs_attrs__}
+        namesmap = {}
+
+        for attribute in type_.__attrs_attrs__:
+            if type(attribute.default) is NOTHING and attribute.init:
+                necessary_fields.add(attribute.name)
+
+            # Manage name mangling
+            if l.mangle_key in attribute.metadata:
+                namesmap[attribute.metadata[l.mangle_key]] = attribute.name
+        l._objfieldscache[type_] = fields, necessary_fields, type_hints, namesmap
+
+    if namesmap:
+        try:
+            value = _mangle_names(namesmap, value, l.failonextra)
+        except ValueError as e:
+            raise TypedloadValueError(str(e), value=value, type_=type_)
+        except AttributeError as e:
+            raise TypedloadAttributeError(str(e), value=value, type_=type_)
+
+    return _objloader(l, fields, necessary_fields, type_hints, value, type_, False)
 
 
 def _strconstructload(l: Loader, value, type_):
@@ -910,11 +952,10 @@ def _iterload(l: Loader, value: Any, type_, function) -> Any:
     t = type_.__args__[0]
 
     # Get function pointer for the handler
-    cached_f = l._indexcache.get(t)
 
-    if cached_f:
-        f = cached_f
-    else:
+    try:
+        f = l._indexcache[t]
+    except KeyError:
         try:
             f = l._indexcache[t] = l.handlers[l.index(t)][1]
         except ValueError:
@@ -928,8 +969,17 @@ def _iterload(l: Loader, value: Any, type_, function) -> Any:
     try:
         ctr = count(1)
         if t in l.basictypes:
+            if function is list:
+                # Copy paste but a list comprehension directly instead of a generic one
+                return [i if isinstance(i, t) else f(l, i, t) for i in compress(value, ctr)]
+            elif function is set:
+                return {i if isinstance(i, t) else f(l, i, t) for i in compress(value, ctr)}
             return function((i if isinstance(i, t) else f(l, i, t) for i in compress(value, ctr)))
         elif is_union(t) and (types := set(uniontypes(t))).issubset(l.basictypes):
+            if function is list:
+                return [i if type(i) in types else f(l, i, t) for i in compress(value, ctr)]
+            elif function is set:
+                return {i if type(i) in types else f(l, i, t) for i in compress(value, ctr)}
             return function((i if type(i) in types else f(l, i, t) for i in compress(value, ctr)))
         else:
             return function(map(f, repeat(l), compress(value, ctr), repeat(t)))
